from .llm_models import _get_model
from .prompts import (
    get_elem_children_instructions,
    get_elem_instructions,
    get_relationships_instructions,
    prompt_template_local,
    schema_to_ttl_instructions,
    ttl_example,
    ttl_to_user_prompt,
)

__all__ = [
    "_get_model",
    "get_elem_instructions",
    "get_elem_children_instructions",
    "get_relationships_instructions",
    "schema_to_ttl_instructions",
    "ttl_example",
    "prompt_template_local",
    "ttl_to_user_prompt",
]



import logging

# Create custom log level
EURAC_LEVEL = 25
logging.addLevelName(EURAC_LEVEL, "EURAC")


def eurac(self, message, *args, **kwargs):
    """
    Log with custom EURAC level
    """
    if self.isEnabledFor(EURAC_LEVEL):
        self._log(EURAC_LEVEL, message, args, **kwargs)


# Add eurac method to Logger class
logging.Logger.eurac = eurac


# Create and configure logger
def get_logger(name="BrickLLM"):
    logger = logging.getLogger(name)

    # Create handler if none exists
    if not logger.handlers:
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    logger.setLevel(EURAC_LEVEL)
    return logger


# Create default logger instance
custom_logger = get_logger()



from typing import Any, Dict

from langchain_core.messages import HumanMessage, SystemMessage

from .. import ElemListSchema, State
from ..helpers import get_elem_children_instructions
from ..logger import custom_logger
from ..utils import create_hierarchical_dict, filter_elements, get_children_hierarchy


def get_elem_children(state: State, config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Identify child elements for each category in the element list using a language model.

    Args:
        state (State): The current state containing the user prompt and element list.
        config (dict): Configuration dictionary containing the language model.

    Returns:
        dict: A dictionary containing the hierarchical structure of identified elements.
    """
    custom_logger.eurac(
        "📊 Getting children for each BrickSchema category in the element list"
    )

    user_prompt = state["user_prompt"]
    categories = state["elem_list"]

    category_dict = {}
    for category in categories:
        children_list = get_children_hierarchy(category, flatten=True)
        children_string = "\n".join(
            [
                f"{parent} -> {child}"
                for parent, child in children_list
                if isinstance(parent, str) and isinstance(child, str)
            ]
        )
        category_dict[category] = children_string

    # Get the model name from the config
    llm = config.get("configurable", {}).get("llm_model")

    # Enforce structured output
    structured_llm = llm.with_structured_output(ElemListSchema)

    identified_children = []
    for category in categories:
        # if the category is not "\n", then add the category to the prompt
        if category_dict[category] != "":
            # System message
            system_message = get_elem_children_instructions.format(
                prompt=user_prompt, elements_list=category_dict[category]
            )
            # Generate question
            elements = structured_llm.invoke(
                [SystemMessage(content=system_message)]
                + [HumanMessage(content="Find the elements.")]
            )
            identified_children.extend(elements.elem_list)
        else:
            identified_children.append(category)

    # Remove duplicates
    identified_children = list(set(identified_children))
    filtered_children = filter_elements(identified_children)

    # create hierarchical dictionary
    hierarchical_dict = create_hierarchical_dict(filtered_children, properties=True)

    return {"elem_hierarchy": hierarchical_dict}



from typing import Any, Dict, List, Union

from langchain.chat_models.base import BaseChatModel
from langgraph.graph import END, START, StateGraph

from .. import GraphConfig, State
from ..edges import validate_condition
from ..nodes import (
    get_elem_children,
    get_elements,
    get_relationships,
    get_sensors,
    schema_to_ttl,
    validate_schema,
)
from .abstract_graph import AbstractBrickSchemaGraph


class BrickSchemaGraph(AbstractBrickSchemaGraph):
    def __init__(self, model: Union[str, BaseChatModel] = "openai"):
        super().__init__(model)
        self.build_graph()
        self.compile_graph()

    def build_graph(self):
        self.workflow = StateGraph(State, config_schema=GraphConfig)

        # Build graph by adding nodes
        self.workflow.add_node("get_elements", get_elements)
        self.workflow.add_node("get_elem_children", get_elem_children)
        self.workflow.add_node("get_relationships", get_relationships)
        self.workflow.add_node("schema_to_ttl", schema_to_ttl)
        self.workflow.add_node("validate_schema", validate_schema)
        self.workflow.add_node("get_sensors", get_sensors)

        # Add edges to define the flow logic
        self.workflow.add_edge(START, "get_elements")
        self.workflow.add_edge("get_elements", "get_elem_children")
        self.workflow.add_edge("get_elem_children", "get_relationships")
        self.workflow.add_edge("get_relationships", "schema_to_ttl")
        self.workflow.add_edge("schema_to_ttl", "validate_schema")
        self.workflow.add_conditional_edges("validate_schema", validate_condition)
        self.workflow.add_edge("get_relationships", "get_sensors")
        self.workflow.add_edge("get_sensors", END)

    def run(
        self, input_data: Dict[str, Any], stream: bool = False
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        if "user_prompt" not in input_data:
            raise ValueError("Input data must contain a 'user_prompt' key.")

        if stream:
            events = []
            for event in self.graph.stream(
                input_data, self.config, stream_mode="values"
            ):
                events.append(event)
            self.result = events[-1]
            self.ttl_output = self.result.get("ttl_output", None)
            return events
        else:
            self.result = self.graph.invoke(input_data, self.config)
            self.ttl_output = self.result.get("ttl_output", None)
            return self.result



"""
Module containing the prompts used for the LLM models
"""

get_elem_instructions: str = """
    You are a BrickSchema ontology expert and you are provided with a user prompt which describes a building or facility.\n
    You are provided with a list of common elements that can be used to describe a building or facility.\n
    You are also provided with the elements description to understand what each element represents.\n
    You are now asked to identify the elements presents in the user prompt, even if not explicitly mentioned.\n
    USER PROMPT: {prompt} \n
    ELEMENTS: {elements_dict} \n
    """  # noqa

get_elem_children_instructions: str = """
    You are a BrickSchema ontology expert and you are provided with a user prompt which describes a building or facility.\n
    You are provided with a list of common elements that can be used to describe a building or facility.\n
    You are now asked to identify the elements presents in the user prompt.\n
    The elements provided are in the format of a hierarchy,
    eg: `Sensor -> Position_Sensor, Sensor -> Energy_Sensor`\n
    You must include only the elements in the list of common elements provided.\n
    DO NOT repeat any elements and DO NOT include "->" in your response.\n

    USER PROMPT: {prompt} \n
    ELEMENTS HIERARCHY: {elements_list} \n
    """  # noqa

get_relationships_instructions: str = """
    You are a BrickSchema ontology expert and are provided with a detailed description of a building or facility.\n
    You are also provided with a hierarchical structure of identified building components.\n
    Your task is to determine the relationships between these components based on the context within the building description and the provided hierarchical structure.\n
    The relationships should reflect direct connections or associations as described or implied in the prompt.\n
    Each element must be followed by a dot symbol (.) and a number to differentiate between elements of the same type (e.g., Room.1, Room.2).\n
    An example of output is the following: [('Building.1', 'Floor.1'), ('Floor.1', 'Room.1'), ('Building.1','Floor.2'), ...]\n
    DO NOT add relationships on the output but only the components names, always add first the parent and then the child.\n
    If an element has no relationships, add an empty string in place of the missing component ("Room.1","").\n
    Hierarchical structure: {building_structure}\n
    USER PROMPT: {prompt}
"""  # noqa

ttl_example: str = """
    @prefix bldg: <urn:Building#> .
    @prefix brick: <https://brickschema.org/schema/Brick#> .
    @prefix prj: <http://example.com/Project#> .

    bldg:CO_sensor a brick:CO ;
        brick:hasTag bldg:hour ;
        brick:hasUnit bldg:PPM ;
        brick:isPointOf bldg: ;
        brick:timeseries [ brick:hasTimeseriesId bldg:jkj4432uz43 ;
                brick:storedAt bldg:example_DB ] .

    bldg:Indoor_humidity a brick:Relative_Humidity_Sensor ;
        brick:hasTag bldg:hour ;
        brick:hasUnit bldg:PERCENT ;
        brick:isPointOf bldg:livingroom ;
        brick:timeseries [ brick:hasTimeseriesId bldg:hfrt56478 ;
                brick:storedAt bldg:example_DB ] .

    bldg:Indoor_temperature a brick:Air_Temperature_Sensor ;
        brick:hasTag bldg:hour ;
        brick:hasUnit bldg:DEG_C ;
        brick:isPointOf bldg:livingroom ;
        brick:timeseries [ brick:hasTimeseriesId bldg:rtg456789 ;
                brick:storedAt bldg:example_DB ] .

    bldg:external_temperature a brick:Air_Temperature_Sensor ;
        brick:hasTag bldg:hour ;
        brick:hasUnit bldg:DEG_C ;
        brick:isPointOf bldg:livingroom ;
        brick:timeseries [ brick:hasTimeseriesId bldg:art53678 ;
                brick:storedAt bldg:example_DB ] .

    bldg:example_db a brick:Database .

    prj:ThermoIot a brick:Site .

    bldg:Milano_Residence_1 a brick:Building ;
        brick:buildingPrimaryFunction [ brick:value "Residential" ] ;
        brick:hasLocation [ brick:value "Milano" ] ;
        brick:isPartOf prj:ThermoIot .

    bldg: a brick:Room ;
        brick:isPartOf bldg:Milano_Residence_1 .

    bldg:livingroom a brick:Room ;
        brick:isPartOf bldg:Milano_Residence_1 .
"""  # noqa

schema_to_ttl_instructions: str = """
    You are a BrickSchema ontology expert and you are provided with a user prompt which describes a building or facility.\n
    You are provided with a dictionary containing the detected components in the building description.\n
    You are also provided with the hierarchical structure of the building components with their constraints BrickSchema compliant.\n
    Your task is to generate a valid TTL (turtle) script that captures the hierarchy and relationships described in the input.\n
    DO NOT add information that are not present in the input.\n
    DO NOT add uuids or database id in the TTL if not specified in the prompt.\n
    You must keep the enumeration with the hashtag for each component otherwise it will not be possible to recognize the components.\n
    The TTL SCRIPT EXAMPLE is useful to understand the overall structure of the output, not the actual content.\n
    TTL SCRIPT EXAMPLE: {ttl_example}\n

    COMPONENTS HIERARCHY: {elem_hierarchy}\n

    USER DESCRIPTION: {prompt}\n

    COMPONENTS DICT: {sensors_dict}\n
"""  # noqa

ttl_to_user_prompt: str = """
    You are a BrickSchema ontology expert tasked with generating a clear and concise description of a building or facility from a TTL script.

    Your output must follow these guidelines:
    - Focus on the key building characteristics, components and relationships present in the TTL
    - Maintain technical accuracy and use proper Brick terminology
    - Keep descriptions clear and well-structured
    - Only include information explicitly stated in the TTL script
    - If no TTL content is provided, return an empty string

    Eventually, the user can provide additional instructions to help you generate the building description.
    <additional_instructions>
    {additional_instructions}
    </additional_instructions>

    TTL script to analyze:
    <ttl_script>
    {ttl_script}
    </ttl_script>
"""  # noqa

prompt_template_local: str = """
    Below is an instruction that describes a task, paired with an input that provides further context. Write a response that appropriately completes the request.

    {instructions}

    ### Input:
    {user_prompt}

    ### Response:
"""  # noqa



from .abstract_graph import AbstractBrickSchemaGraph
from .brickschema_graph import BrickSchemaGraph
from .brickschema_graph_local import BrickSchemaGraphLocal

__all__ = ["AbstractBrickSchemaGraph", "BrickSchemaGraph", "BrickSchemaGraphLocal"]



from typing import Union

from langchain.chat_models.base import BaseChatModel
from langchain_anthropic import ChatAnthropic
from langchain_community.llms import Ollama
from langchain_fireworks import ChatFireworks
from langchain_openai import ChatOpenAI


def _get_model(model: Union[str, BaseChatModel]) -> BaseChatModel:
    """
    Get the LLM model based on the provided model type.

    Args:
        model (Union[str, BaseChatModel]): The model type as a string or an instance of BaseChatModel.

    Returns:
        BaseChatModel: The LLM model instance.
    """

    if isinstance(model, BaseChatModel):
        return model

    if model == "openai":
        return ChatOpenAI(temperature=0, model="gpt-4o")
    elif model == "anthropic":
        return ChatAnthropic(temperature=0, model_name="claude-3-sonnet-20240229")
    elif model == "fireworks":
        return ChatFireworks(
            temperature=0, model="accounts/fireworks/models/llama-v3p1-70b-instruct"
        )
    elif model == "llama3.1:8b-brick":
        return Ollama(model="llama3.1:8b-brick-v8")
    else:
        raise ValueError(f"Unsupported model type: {model}")



from .configs import GraphConfig
from .logger import custom_logger
from .schemas import (
    ElemListSchema,
    RelationshipsSchema,
    TTLSchema,
    TTLToBuildingPromptSchema,
)
from .states import State, StateLocal

__all__ = [
    "ElemListSchema",
    "RelationshipsSchema",
    "TTLSchema",
    "TTLToBuildingPromptSchema",
    "State",
    "StateLocal",
    "GraphConfig",
    "custom_logger",
]



from .get_hierarchy_info import (
    build_hierarchy,
    create_hierarchical_dict,
    extract_ttl_content,
    filter_elements,
    find_parents,
    find_sensor_paths,
    flatten_hierarchy,
    get_all_subchildren,
    get_children,
    get_children_hierarchy,
    get_hierarchical_info,
)
from .query_brickschema import (
    clean_result,
    general_query,
    get_brick_definition,
    get_query_result,
    iterative_subclasses,
    query_properties,
    query_subclass,
    validate_ttl,
)
from .rdf_parser import extract_rdf_graph
from .ttl_to_prompt import ttl_to_building_prompt

__all__ = [
    "find_parents",
    "get_children",
    "flatten_hierarchy",
    "get_hierarchical_info",
    "get_all_subchildren",
    "get_children_hierarchy",
    "filter_elements",
    "create_hierarchical_dict",
    "find_sensor_paths",
    "build_hierarchy",
    "extract_ttl_content",
    "get_brick_definition",
    "get_query_result",
    "clean_result",
    "query_subclass",
    "query_properties",
    "iterative_subclasses",
    "general_query",
    "validate_ttl",
    "extract_rdf_graph",
    "ttl_to_building_prompt",
]



from typing import Any, Dict

from .. import State
from ..logger import custom_logger


def get_sensors(state: State) -> Dict[str, Any]:
    """
    Retrieve sensor information for the building structure.

    Args:
        state (State): The current state.

    Returns:
        dict: A dictionary containing sensor UUIDs mapped to their locations.
    """
    custom_logger.eurac("📡 Getting sensors information")

    uuid_dict = {
        "Building#1>Floor#1>Office#1>Room#1": [
            {
                "name": "Temperature_Sensor#1",
                "uuid": "aaaa-bbbb-cccc-dddd",
            },
            {
                "name": "Humidity_Sensor#1",
                "uuid": "aaaa-bbbb-cccc-dddd",
            },
        ],
        "Building#1>Floor#1>Office#1>Room#2": [
            {
                "name": "Temperature_Sensor#2",
                "uuid": "aaaa-bbbb-cccc-dddd",
            },
            {
                "name": "Humidity_Sensor#2",
                "uuid": "aaaa-bbbb-cccc-dddd",
            },
        ],
    }
    return {"uuid_dict": uuid_dict}



from typing import Any, Dict, List

from typing_extensions import TypedDict


# state for BrickSchemaGraph class
class State(TypedDict):
    user_prompt: str
    elem_list: List[str]
    # elem_children_list: List[str]
    elem_hierarchy: Dict[str, Any]
    # relationships: List[Tuple[str, str]]
    # rel_tree: Dict[str, Any]
    sensors_dict: Dict[str, List[str]]
    is_valid: bool
    validation_report: str
    validation_max_iter: int
    uuid_dict: Dict[str, Any]
    ttl_output: str


# state for BrickSchemaGraphLocal class
class StateLocal(TypedDict):
    instructions: str
    user_prompt: str
    is_valid: bool
    validation_report: str
    validation_max_iter: int
    ttl_output: str



import logging
from typing import List, Optional, Tuple, Union

from langchain.chat_models.base import BaseChatModel
from langchain_community.llms import Ollama
from langchain_core.messages import HumanMessage, SystemMessage

from ..helpers import ttl_to_user_prompt
from ..schemas import TTLToBuildingPromptSchema

logger = logging.getLogger(__name__)


def ttl_to_building_prompt(
    ttl_file: str,
    llm: Union[Ollama, BaseChatModel],
    additional_instructions: Optional[str] = None,
) -> Tuple[str, List[str]]:

    # Enforce structured output
    structured_llm = llm.with_structured_output(TTLToBuildingPromptSchema)

    # System message
    system_message = ttl_to_user_prompt.format(
        ttl_script=ttl_file, additional_instructions=additional_instructions
    )

    logger.info("Generating building description and key elements from the TTL file.")
    # Generate question
    answer = structured_llm.invoke(
        [SystemMessage(content=system_message)]
        + [HumanMessage(content="Generate the TTL.")]
    )

    return answer.building_description, answer.key_elements



from typing import List, Tuple

from pydantic.v1 import BaseModel, Field


# pydantic schemas
class ElemListSchema(BaseModel):
    elem_list: List[str]


class RelationshipsSchema(BaseModel):
    relationships: List[Tuple[str, ...]]


class TTLSchema(BaseModel):
    ttl_output: str = Field(
        ..., description="The generated BrickSchema turtle/rdf script."
    )


class TTLToBuildingPromptSchema(BaseModel):
    building_description: str = Field(
        ..., description="The generated building description."
    )
    key_elements: List[str] = Field(
        ..., description="The generated list of key elements."
    )



import os
import re
from io import StringIO
from typing import Dict, List, Optional, Tuple, Union

import pkg_resources
import pyshacl
from rdflib import Graph, Namespace, URIRef
from rdflib.query import ResultRow

# Path to the Brick schema Turtle file
brick_ttl_path = pkg_resources.resource_filename(
    __name__, os.path.join("..", "ontologies", "Brick.ttl")
)
# Load the Brick schema Turtle file
g = Graph()
g.parse(brick_ttl_path, format="ttl")

# Define the namespaces from the prefixes
namespaces = {
    "brick": Namespace("https://brickschema.org/schema/Brick#"),
}


# Function to get the definition from the TTL file
def get_brick_definition(element_name: str) -> str:
    """
    Get the definition of an element from the Brick schema Turtle file.

    Args:
        element_name (str): The name of the element to get the definition for.

    Returns:
        str: The definition of the element, or "No definition available" if not found.
    """
    normalized_key = element_name.replace("_", "").lower()
    for prefix, namespace in namespaces.items():
        uri = namespace[element_name]
        for s, p, o in g.triples(
            (uri, URIRef("http://www.w3.org/2004/02/skos/core#definition"), None)
        ):
            return str(o)
        uri = namespace[normalized_key]
        for s, p, o in g.triples(
            (uri, URIRef("http://www.w3.org/2004/02/skos/core#definition"), None)
        ):
            return str(o)
    return "No definition available"


# Function to get the query result without using pandas
def get_query_result(query: str) -> List[Dict[str, str]]:
    """
    Execute a SPARQL query on the Brick schema graph and return the results.

    Args:
        query (str): The SPARQL query to execute.

    Returns:
        List[Dict[str, str]]: A list of dictionaries representing the query results.
    """
    result = g.query(query)
    # Convert the result to a list of dictionaries where keys are the variable names
    query_vars = list(result.vars) if result.vars is not None else []
    data: List[Dict[str, Optional[str]]] = []
    for row in result:
        if isinstance(row, ResultRow):
            data.append(
                {str(var): str(row[var]) if row[var] else None for var in query_vars}
            )
    # Remove entries with None values and reset index
    cleaned_data = [
        {key: value for key, value in row.items() if value is not None} for row in data
    ]
    return cleaned_data


# Function to clean the result, extracting the needed part of the URI
def clean_result(data: List[str]) -> List[str]:
    """
    Extract the relevant part of a URI from a list of data.

    Args:
        data (List[str]): A list of URIs to clean.

    Returns:
        List[str]: A list of extracted parts from the URIs.
    """
    return [
        re.findall(r"#(\w+)", value)[0]
        for value in data
        if re.findall(r"#(\w+)", value)
    ]


# Function to create a SPARQL query for subclasses
def query_subclass(element: str) -> str:
    """
    Create a SPARQL query to find subclasses of a given element.

    Args:
        element (str): The element to find subclasses for.

    Returns:
        str: The SPARQL query string.
    """
    return f"SELECT ?subclass WHERE {{ brick:{element} rdfs:subClassOf ?subclass . }}"


# Function to create a SPARQL query for properties
def query_properties(element: str) -> str:
    """
    Create a SPARQL query to find properties of a given element.

    Args:
        element (str): The element to find properties for.

    Returns:
        str: The SPARQL query string.
    """
    return f"""
    SELECT ?property ?message ?path ?class WHERE {{
        brick:{element} sh:property ?property .
        ?property sh:message ?message ; sh:path ?path ;
                  sh:or/rdf:rest*/rdf:first ?constraint .
        ?constraint sh:class ?class .
    }}
    """


# Function to iteratively find subclasses
def iterative_subclasses(element: str) -> List[str]:
    """
    Iteratively find all subclasses of a given element.

    Args:
        element (str): The element to find subclasses for.

    Returns:
        List[str]: A list of subclasses.
    """
    subclasses: List[str] = []
    sub_class_data = get_query_result(query_subclass(element))
    subClass = (
        clean_result([row["subclass"] for row in sub_class_data])
        if sub_class_data
        else []
    )

    while subClass:
        subclasses.append(subClass[0])
        if subClass[0] in {
            "Collection",
            "Equipment",
            "Location",
            "Measureable",
            "Point",
        }:
            break
        sub_class_data = get_query_result(query_subclass(subClass[0]))
        subClass = (
            clean_result([row["subclass"] for row in sub_class_data])
            if sub_class_data
            else []
        )

    return subclasses


# General query function to retrieve properties and relationships
def general_query(element: str) -> Dict[str, Dict[str, Union[str, List[str]]]]:
    """
    Retrieve properties and relationships for a given element.

    Args:
        element (str): The element to retrieve properties and relationships for.

    Returns:
        Dict[str, Dict[str, Union[str, List[str]]]]: A dictionary containing properties and their constraints.
    """
    subclasses = iterative_subclasses(element)
    if not subclasses:
        return {}

    query_data = get_query_result(query_properties(subclasses[-1]))
    relationships: Dict[str, Dict[str, Union[str, List[str]]]] = {}

    for row in query_data:
        property_name = clean_result([row["path"]])[0]
        if property_name not in relationships:
            relationships[property_name] = {
                "message": row["message"],
                "constraint": clean_result([row["class"]]),
            }
        else:
            if isinstance(relationships[property_name]["constraint"], list):
                relationships[property_name]["constraint"].extend(
                    clean_result([row["class"]])
                )
            else:
                relationships[property_name]["constraint"] = clean_result(
                    [row["class"]]
                )

    return {"property": relationships}


def validate_ttl(ttl_file: str, method: str = "pyshacl") -> Tuple[bool, str]:
    """
    Validate a TTL file using the specified method.

    Args:
        ttl_file (str): The TTL file to validate.
        method (str): The method to use for validation. Default is 'pyshacl'.

    Returns:
        Tuple[bool, str]: A tuple containing a boolean indicating if the validation was successful and a validation report or error message.
    """
    # Load the ttl file
    output_graph = Graph()
    try:
        output_graph.parse(StringIO(ttl_file), format="ttl")
    except Exception as e:
        return False, f"Failed to parse the TTL file. Content: {e}"

    if method == "pyshacl":
        valid, results_graph, report = pyshacl.validate(
            output_graph,
            shacl_graph=g,
            ont_graph=g,
            inference="rdfs",
            abort_on_first=False,
            allow_infos=False,
            allow_warnings=False,
            meta_shacl=False,
            advanced=False,
            js=False,
            debug=False,
        )
        return valid, report
    else:
        return False, "Method not found"



from .generation_local import generation_local
from .get_elem_children import get_elem_children
from .get_elements import get_elements
from .get_relationships import get_relationships
from .get_sensors import get_sensors
from .schema_to_ttl import schema_to_ttl
from .validate_schema import validate_schema

__all__ = [
    "get_elem_children",
    "get_elements",
    "get_relationships",
    "get_sensors",
    "schema_to_ttl",
    "validate_schema",
    "generation_local",
]



from typing import Literal, TypedDict, Union

from langchain.chat_models.base import BaseChatModel


# Define the config
class GraphConfig(TypedDict):
    model: Union[
        Literal["anthropic", "openai", "fireworks", "llama3.1:8b-brick"], BaseChatModel
    ]



from typing import Any, Dict

from ..logger import custom_logger
from ..utils import validate_ttl


def validate_schema(state) -> Dict[str, Any]:
    """
    Validate the generated TTL output against the BrickSchema.

    Args:
        state (State): The current state containing the TTL output and validation parameters.

    Returns:
        dict: A dictionary containing the validation status and report.
    """
    custom_logger.eurac("✅ Validating TTL schema")

    ttl_output = state.get("ttl_output", None)
    max_iter = state.get("validation_max_iter", 2)

    max_iter -= 1

    if ttl_output is None:
        return {
            "is_valid": False,
            "validation_report": "Empty TTL output.",
            "validation_max_iter": max_iter,
        }

    is_valid, report = validate_ttl(ttl_output)

    return {
        "is_valid": is_valid,
        "validation_report": report,
        "validation_max_iter": max_iter,
    }



from .graphs import BrickSchemaGraph, BrickSchemaGraphLocal

# compiled graph
brickschema_graph = BrickSchemaGraph()._compiled_graph()
brickschema_graph_local = BrickSchemaGraphLocal()._compiled_graph()



import re


def extract_rdf_graph(llm_response: str) -> str:

    all_lines = llm_response.splitlines()
    for i, line in enumerate(all_lines):
        all_lines[i] = line.strip()
    llm_response = "\n".join(all_lines).strip()

    if not llm_response.strip().startswith("@prefix "):
        # Try to find RDF content within backticks
        backtick_pattern = re.compile(r"```(.*?)```", re.DOTALL)
        match = backtick_pattern.search(llm_response)
        if match:
            llm_response = match.group(1).strip()
            # return rdf_graph
        # If no backticks, look for RDF starting with @prefix
        rdf_start_pattern = re.compile(r"@prefix [^\s]*: <[^>]*", re.DOTALL)
        match = rdf_start_pattern.search(llm_response)
        if match:
            start_index = match.start()
            rdf_content = llm_response[start_index:].strip()
        else:
            # If no valid RDF content is found, raise an error
            raise ValueError("No valid RDF found in the provided graph content.")
    else:
        rdf_content = llm_response.strip()
    lines = rdf_content.splitlines()
    if lines and lines[-1].strip().endswith("```"):
        rdf_graph = "\n".join(lines[:-1])  # Remove the last line
    else:
        flag_last_line = False
        while not flag_last_line:
            last_line = lines[-1].strip() if lines else ""
            if any(
                word in last_line
                for word in ["note", "Note", "Please", "please", "Here", "here"]
            ):
                lines.pop()
                # TODO: Extract user namespace from instructions and insert instead of bldg:
            elif not (
                last_line.startswith("bldg:")
                or last_line.startswith("ref:")
                or last_line.startswith("unit:")
                or last_line.startswith("brick:")
                or last_line.startswith("a")
            ):  # or any(["note", "Note", "Please"]) in last_line:
                lines.pop()
            elif "```" in last_line.strip():
                lines.pop()
            else:
                flag_last_line = True
        rdf_graph = "\n".join(lines).strip()

        return rdf_graph



import json
from typing import Any, Dict

from langchain_core.messages import HumanMessage, SystemMessage

from .. import State, TTLSchema
from ..helpers import schema_to_ttl_instructions, ttl_example
from ..logger import custom_logger


def schema_to_ttl(state: State, config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Generate a TTL (Turtle) script from the building description and component hierarchy.

    Args:
        state (State): The current state containing the user prompt, sensors, and element hierarchy.
        config (dict): Configuration dictionary containing the language model.

    Returns:
        dict: A dictionary containing the generated TTL output.
    """
    custom_logger.eurac("📝 Generating TTL from schema")

    user_prompt = state["user_prompt"]
    sensors_dict = state["sensors_dict"]
    elem_hierarchy = state["elem_hierarchy"]

    sensors_dict_json = json.dumps(sensors_dict, indent=2)
    elem_hierarchy_json = json.dumps(elem_hierarchy, indent=2)

    # Get the model name from the config
    llm = config.get("configurable", {}).get("llm_model")

    # Enforce structured output
    structured_llm = llm.with_structured_output(TTLSchema)

    # System message
    system_message = schema_to_ttl_instructions.format(
        prompt=user_prompt,
        sensors_dict=sensors_dict_json,
        elem_hierarchy=elem_hierarchy_json,
        ttl_example=ttl_example,
    )

    # Generate question
    answer = structured_llm.invoke(
        [SystemMessage(content=system_message)]
        + [HumanMessage(content="Generate the TTL.")]
    )

    return {"ttl_output": answer.ttl_output}



from typing import Any, Dict

from langchain_core.messages import HumanMessage, SystemMessage

from .. import ElemListSchema, State
from ..helpers import get_elem_instructions
from ..logger import custom_logger
from ..utils import get_brick_definition, get_hierarchical_info


def get_elements(state: State, config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Process the user prompt to identify elements within specified categories
    using a language model.

    Args:
        state (State): The current state containing the user prompt.
        config (dict): Configuration dictionary containing the language model.

    Returns:
        dict: A dictionary containing the list of identified elements.
    """
    custom_logger.eurac("🔍 Getting elements from user prompt")

    user_prompt = state["user_prompt"]

    categories = ["Point", "Equipment", "Location", "Collection"]

    category_dict = {}
    # Get hierarchy info for each category
    for category in categories:
        parents, children = get_hierarchical_info(category)
        # category_dict[category] = children

        # get definition for each child
        children_dict = {}
        for child in children:
            children_dict[child] = get_brick_definition(child)

        category_dict[category] = children_dict

    # Get the model name from the config
    llm = config.get("configurable", {}).get("llm_model")

    # Enforce structured output
    structured_llm = llm.with_structured_output(ElemListSchema)

    # System message
    system_message = get_elem_instructions.format(
        prompt=user_prompt, elements_dict=category_dict
    )

    # Generate question
    answer = structured_llm.invoke(
        [SystemMessage(content=system_message)]
        + [HumanMessage(content="Find the elements.")]
    )

    return {"elem_list": answer.elem_list}



from typing import Any, Dict, List, Union

from langchain.chat_models.base import BaseChatModel
from langgraph.graph import START, StateGraph

from .. import GraphConfig, StateLocal
from ..edges import validate_condition_local
from ..nodes import generation_local, validate_schema
from .abstract_graph import AbstractBrickSchemaGraph


class BrickSchemaGraphLocal(AbstractBrickSchemaGraph):
    def __init__(self, model: Union[str, BaseChatModel]):
        super().__init__(model)
        self.build_graph()
        self.compile_graph()

    def build_graph(self):
        self.workflow = StateGraph(StateLocal, config_schema=GraphConfig)

        # Build graph by adding nodes
        self.workflow.add_node("generation_local", generation_local)
        self.workflow.add_node("validate_schema", validate_schema)

        # Add edges to define the flow logic
        self.workflow.add_edge(START, "generation_local")
        self.workflow.add_edge("generation_local", "validate_schema")
        self.workflow.add_conditional_edges("validate_schema", validate_condition_local)

    def run(
        self, input_data: Dict[str, Any], stream: bool = False
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        if "user_prompt" not in input_data or "instructions" not in input_data:
            raise ValueError(
                "Input data must contain both 'user_prompt' and 'instructions' keys."
            )

        if stream:
            events = []
            for event in self.graph.stream(
                input_data, self.config, stream_mode="values"
            ):
                events.append(event)
            self.result = events[-1]
            self.ttl_output = self.result.get("ttl_output", None)
            return events
        else:
            self.result = self.graph.invoke(input_data, self.config)
            self.ttl_output = self.result.get("ttl_output", None)
            return self.result



from typing import Any, Dict

from .. import StateLocal
from ..helpers import prompt_template_local
from ..logger import custom_logger
from ..utils import extract_rdf_graph


def generation_local(state: StateLocal, config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Generate the RDF graph using the local LLM from a prompt containing the building description and the instruction

    Args:
        state (StateLocal): The current state containing the user prompt and the instructions.
        config (dict): Configuration dictionary containing the language model.

    Returns:
        dict: A dictionary containing the output generated.
    """

    custom_logger.eurac("🤖 Starting one shot generation with local LLM")

    instructions = state["instructions"]
    user_prompt = state["user_prompt"]

    llm = config.get("configurable", {}.get("llm_model"))["llm_model"]

    message = prompt_template_local.format(
        instructions=instructions, user_prompt=user_prompt
    )

    answer = llm.invoke(message)
    ttl_output = extract_rdf_graph(answer)

    return {"ttl_output": ttl_output}



import os
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple, Union

from langchain.chat_models.base import BaseChatModel
from langgraph.graph import StateGraph
from PIL import Image

from ..helpers.llm_models import _get_model
from ..utils import ttl_to_building_prompt


class AbstractBrickSchemaGraph(ABC):
    def __init__(self, model: Union[str, BaseChatModel] = "openai"):
        self.model = _get_model(model)
        self.workflow = None
        self.graph = None
        self.config = {"configurable": {"thread_id": "1", "llm_model": self.model}}
        self.result = None
        self.ttl_output = None
        self.generated_building_description = None
        self.generated_key_elements = None

    @abstractmethod
    def build_graph(self):
        """Build the graph by adding nodes and edges."""
        pass

    def compile_graph(self):
        """Compile the graph."""
        try:
            self.graph = self.workflow.compile()
        except Exception as e:
            raise ValueError(f"Failed to compile the graph: {e}")

    def _compiled_graph(self) -> StateGraph:
        """Check if the graph is compiled and return the compiled graph."""
        if self.graph is None:
            raise ValueError(
                "Graph is not compiled yet. Please compile the graph first."
            )
        return self.graph

    @abstractmethod
    def run(
        self, input_data: Dict[str, Any], stream: bool = False
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """Run the graph with the given input data."""
        pass

    def display(self, filename: str = "graph.png") -> None:
        """Display the compiled graph as an image."""
        if self.graph is None:
            raise ValueError(
                "Graph is not compiled yet. Please compile the graph first."
            )

        # Save the image to the specified file
        self.graph.get_graph().draw_mermaid_png(output_file_path=filename)

        # Open the image using PIL (Pillow)
        if os.path.exists(filename):
            with Image.open(filename) as img:
                img.show()
        else:
            raise FileNotFoundError(
                f"Failed to generate the graph image file: {filename}"
            )

    def get_state_snapshots(self) -> List[Dict[str, Any]]:
        """Get all the state snapshots from the graph execution."""
        all_states = []
        for state in self.graph.get_state_history(self.config):
            all_states.append(state)
        return all_states

    def save_ttl_output(self, output_file: str = "brick_schema_output.ttl") -> None:
        """Save the TTL output to a file."""
        if self.result is None:
            raise ValueError("No result found. Please run the graph first.")

        if self.ttl_output is None:
            raise ValueError("No TTL output found. Please run the graph first.")

        with open(output_file, "w") as f:
            f.write(self.ttl_output)

    def ttl_to_building_description(self) -> Tuple[str, List[str]]:
        if self.ttl_output is None:
            raise ValueError("No TTL output found. Please run the graph first.")
        self.generated_building_description, self.generated_key_elements = (
            ttl_to_building_prompt(self.ttl_output, self.model)
        )
        return self.generated_building_description, self.generated_key_elements



import json
import os
import re
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Union

import pkg_resources

from .query_brickschema import general_query

# Path to the JSON file
brick_hierarchy_path = pkg_resources.resource_filename(
    __name__, os.path.join("..", "ontologies", "brick_hierarchy.json")
)

# Load the JSON file
with open(brick_hierarchy_path) as f:
    data = json.load(f)


# Function to recursively find parents
def find_parents(
    current_data: Dict[str, Any], target: str, parents: Optional[List[str]] = None
) -> Tuple[bool, List[str]]:
    """
    Recursively find the parent nodes of a target node in a hierarchical data structure.

    Args:
        current_data (Dict[str, Any]): The current level of the hierarchy to search.
        target (str): The target node to find parents for.
        parents (Optional[List[str]], optional): Accumulated list of parent nodes. Defaults to None.

    Returns:
        Tuple[bool, List[str]]: A tuple containing a boolean indicating if the target was found and a list of parent nodes.
    """
    if parents is None:
        parents = []
    for key, value in current_data.items():
        if key == target:
            return True, parents
        if isinstance(value, dict):
            found, result = find_parents(value, target, parents + [key])
            if found:
                return True, result
    return False, []


# Function to get the children of a node
def get_children(current_data: Dict[str, Any], target: str) -> List[str]:
    """
    Get the children of a target node in a hierarchical data structure.

    Args:
        current_data (Dict[str, Any]): The current level of the hierarchy to search.
        target (str): The target node to find children for.

    Returns:
        List[str]: A list of child nodes.
    """
    if target in current_data:
        return list(current_data[target].keys())
    for key, value in current_data.items():
        if isinstance(value, dict):
            children = get_children(value, target)
            if children:
                return children
    return []


# Function to flatten the hierarchy
def flatten_hierarchy(
    current_data: Dict[str, Any],
    parent: Optional[str] = None,
    result: Optional[List[Tuple[str, str]]] = None,
) -> List[Tuple[str, str]]:
    """
    Flatten a hierarchical data structure into a list of parent-child tuples.

    Args:
        current_data (Dict[str, Any]): The current level of the hierarchy to flatten.
        parent (Optional[str], optional): The parent node. Defaults to None.
        result (Optional[List[Tuple[str, str]]], optional): Accumulated list of parent-child tuples. Defaults to None.

    Returns:
        List[Tuple[str, str]]: A list of tuples representing parent-child relationships.
    """
    if result is None:
        result = []
    for key, value in current_data.items():
        if parent:
            result.append((parent, key))
        if isinstance(value, dict):
            flatten_hierarchy(value, key, result)
    return result


# Main function to get hierarchy info
def get_hierarchical_info(key: str) -> Tuple[List[str], List[str]]:
    """
    Get the hierarchical information of a node, including its parents and children.

    Args:
        key (str): The target node to get information for.

    Returns:
        Tuple[List[str], List[str]]: A tuple containing a list of parent nodes and a list of child nodes.
    """
    # Get parents
    found, parents = find_parents(data, key)
    # Get children
    children = get_children(data, key)
    return (parents, children)


# Function to recursively get all children and subchildren
def get_all_subchildren(current_data: Dict[str, Any], target: str) -> Dict[str, Any]:
    """
    Recursively get all children and subchildren of a target node.

    Args:
        current_data (Dict[str, Any]): The current level of the hierarchy to search.
        target (str): The target node to find children for.

    Returns:
        Dict[str, Any]: A dictionary representing the subtree of the target node.
    """
    if target in current_data:
        sub_tree = current_data[target]
        if isinstance(sub_tree, dict):
            return sub_tree
        else:
            return {}
    for key, value in current_data.items():
        if isinstance(value, dict):
            result = get_all_subchildren(value, target)
            if result:
                return result
    return {}


# Main function to get hierarchy dictionary
def get_children_hierarchy(
    key: str, flatten: bool = False
) -> Union[Dict[str, Any], List[Tuple[str, str]]]:
    """
    Get the hierarchy of children for a target node, optionally flattening the result.

    Args:
        key (str): The target node to get children for.
        flatten (bool, optional): Whether to flatten the hierarchy. Defaults to False.

    Returns:
        Union[Dict[str, Any], List[Tuple[str, str]]]: A dictionary representing the hierarchy or a list of parent-child tuples if flattened.
    """
    if flatten:
        return flatten_hierarchy(get_all_subchildren(data, key))
    return get_all_subchildren(data, key)


# Function to filter elements based on the given conditions
def filter_elements(elements: List[str]) -> List[str]:
    """
    Filter elements based on their hierarchical relationships.

    Args:
        elements (List[str]): A list of elements to filter.

    Returns:
        List[str]: A list of filtered elements.
    """
    elements_info = {element: get_hierarchical_info(element) for element in elements}
    filtered_elements = []

    for element, (parents, children) in elements_info.items():
        # Discard elements with no parents and no children
        if not parents and not children:
            continue
        # Check if the element is a parent of any other element
        is_parent = any(element in p_list for p_list, _ in elements_info.values())
        if is_parent:
            continue
        filtered_elements.append(element)

    return filtered_elements


def create_hierarchical_dict(
    elements: List[str], properties: bool = False
) -> Dict[str, Any]:
    """
    Create a hierarchical dictionary from a list of elements, optionally including properties.

    Args:
        elements (List[str]): A list of elements to include in the hierarchy.
        properties (bool, optional): Whether to include properties in the hierarchy. Defaults to False.

    Returns:
        Dict[str, Any]: A dictionary representing the hierarchical structure.
    """
    hierarchy: Dict[str, Any] = {}

    for category in elements:
        parents, _ = get_hierarchical_info(category)
        current_level = hierarchy

        for parent in parents:
            if parent not in current_level:
                current_level[parent] = {}
            current_level = current_level[parent]

        # Finally add the category itself
        if category not in current_level:
            if properties:
                elem_property = general_query(category)
                if len(elem_property.keys()) == 0:
                    continue
                elem_property = elem_property["property"]
                # remove "message" key from the dictionary
                for prop in elem_property.keys():
                    elem_property[prop].pop("message")
                current_level[category] = {"properties": elem_property}
            else:
                current_level[category] = {}

    return hierarchy


def find_sensor_paths(
    tree: Dict[str, Any], path: Optional[List[str]] = None
) -> List[Dict[str, str]]:
    """
    Find paths to sensor nodes in a hierarchical tree structure.

    Args:
        tree (Dict[str, Any]): The hierarchical tree structure.
        path (Optional[List[str]], optional): Accumulated path to the current node. Defaults to None.

    Returns:
        List[Dict[str, str]]: A list of dictionaries containing sensor names and their paths.
    """
    if path is None:
        path = []

    current_path = path + [tree.get("name", "")]
    if "children" not in tree or not tree["children"]:
        if re.search(r"Sensor", tree.get("name", "")):
            sensor_path = ">".join(current_path[:-1])
            return [{"name": tree.get("name", ""), "path": sensor_path}]
        return []

    sensor_paths = []
    for child in tree["children"]:
        sensor_paths.extend(find_sensor_paths(child, current_path))

    return sensor_paths


def build_hierarchy(relationships: List[Tuple[str, str]]) -> Dict[str, Any]:
    """
    Build a hierarchical tree structure from a list of parent-child relationships.

    Args:
        relationships (List[Tuple[str, str]]): A list of tuples representing parent-child relationships.

    Returns:
        Dict[str, Any]: A dictionary representing the hierarchical tree structure.
    """

    # Helper function to recursively build the tree structure
    def build_tree(node: str, tree_dict: Dict[str, List[str]]) -> Dict[str, Any]:
        return (
            {
                "name": node,
                "children": [build_tree(child, tree_dict) for child in tree_dict[node]],
            }
            if tree_dict[node]
            else {"name": node, "children": []}
        )

    # Create a dictionary to hold parent-children relationships
    tree_dict: Dict[str, List[str]] = defaultdict(list)
    nodes = set()

    # Fill the dictionary with data from relationships
    for parent, child in relationships:
        tree_dict[parent].append(child)
        nodes.update([parent, child])

    # Find the root (a node that is never a child)
    root_candidates = {
        node for node in nodes if node not in {child for _, child in relationships}
    }
    if not root_candidates:
        raise ValueError("No root found in relationships")
    root = next(iter(root_candidates))

    # Build the hierarchical structure starting from the root
    hierarchy = build_tree(root, tree_dict)
    return hierarchy


def extract_ttl_content(input_string: str) -> str:
    """
    Extract content between code block markers in a string.

    Args:
        input_string (str): The input string containing code blocks.

    Returns:
        str: The extracted content between the code block markers.
    """
    # Use regex to match content between ```code and ```
    match = re.search(r"```code\s*(.*?)\s*```", input_string, re.DOTALL)
    if match:
        return match.group(1).strip()
    return ""



import json
from collections import defaultdict
from typing import Any, Dict

from langchain_core.messages import HumanMessage, SystemMessage

from .. import RelationshipsSchema, State
from ..helpers import get_relationships_instructions
from ..logger import custom_logger
from ..utils import build_hierarchy, find_sensor_paths


def get_relationships(state: State, config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Determine relationships between building components using a language model.

    Args:
        state (State): The current state containing the user prompt and element hierarchy.
        config (dict): Configuration dictionary containing the language model.

    Returns:
        dict: A dictionary containing the grouped sensor paths.
    """
    custom_logger.eurac("🔗 Getting relationships between building components")

    user_prompt = state["user_prompt"]
    building_structure = state["elem_hierarchy"]

    # Convert building structure to a JSON string for better readability
    building_structure_json = json.dumps(building_structure, indent=2)

    # Get the model name from the config
    llm = config.get("configurable", {}).get("llm_model")

    # Enforce structured output
    structured_llm = llm.with_structured_output(RelationshipsSchema)
    # System message
    system_message = get_relationships_instructions.format(
        prompt=user_prompt, building_structure=building_structure_json
    )

    # Generate question
    answer = structured_llm.invoke(
        [SystemMessage(content=system_message)]
        + [HumanMessage(content="Find the relationships.")]
    )

    try:
        tree_dict = build_hierarchy(answer.relationships)
    except Exception as e:
        print(f"Error building the hierarchy: {e}")

    # Group sensors by their paths
    sensor_paths = find_sensor_paths(tree_dict)
    grouped_sensors = defaultdict(list)
    for sensor in sensor_paths:
        grouped_sensors[sensor["path"]].append(sensor["name"])
    grouped_sensor_dict = dict(grouped_sensors)

    return {"sensors_dict": grouped_sensor_dict}



from typing import Any, Dict, Literal


def validate_condition(state: Dict[str, Any]) -> Literal["schema_to_ttl", "__end__"]:
    """
    Validate the condition for the next node to visit.

    Args:
        state (Dict[str, Any]): The current state containing the validation result.

    Returns:
        Literal["schema_to_ttl", "__end__"]: The next node to visit.
    """

    is_valid = state.get("is_valid", False)
    max_iter = state.get("validation_max_iter", 0)

    if max_iter > 0 and not is_valid:
        return "schema_to_ttl"

    return "__end__"
